#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <netinet/in.h>
#include <netinet/tcp.h>

#include <sys/types.h>
#include <sys/socket.h>

#include <inttypes.h>

#include <gsloop_types.h>
#include <gsloop_server.h>
#include <gsloop_con.h>


/**
 * make_socket_inet - create a new socket for PF_INET
 */
static int gsloop_inet_server_open(int		type,
				   uint16_t	port)
{
	struct sockaddr_in	name;
	int			new_sock;
	int			yes = 1;
	int			err;

	/* Create the socket.  */
	new_sock = socket(PF_INET, type, 0);
	if (new_sock == -1) {
		perror("socket");
		err = -errno;
		goto exit_0;
	}
	
	err = setsockopt(new_sock, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof (yes));
	if (err < 0) {
		perror ("setsockopt");
		err = -errno;
		goto exit_1;
	}

	/* Give the socket a name */
	memset(&name, 0, sizeof(name));
	name.sin_family	     = type;
	name.sin_port	     = htons(port);
	name.sin_addr.s_addr = htonl(INADDR_ANY);
	err = bind(new_sock, (struct sockaddr *)&name, sizeof(name));
	if (err < 0) {
		perror("bind");
		err = -errno;
		goto exit_1;
	}

	/* we are server, so listen to incoming connections */
	err = listen(new_sock, 8);	/* FIXME: backlog */
	if (err < 0) {
		perror("listen");
		err = -errno;
		goto exit_1;
	}

	return new_sock;

 exit_1:
	close(new_sock);
 exit_0:
	return err;
}


static int gsloop_inet_client_close(union gsloop_con	*con)
{
	struct gsloop_client	*client = &con->client;
	struct gsloop		*gsloop = client->gsloop;
	int			err;

	err = rn_del(&gsloop->rn, client->fd);
	if (err) {
		err = -err;	/* rn returns errno */
		dprintf(LEVEL_ERROR, "(%s) rn_del failed, err %d\n",
			__FUNCTION__, err);
	}

	err = shutdown(client->fd, SHUT_RDWR);
	if (err) {
		err = -errno;
		dprintf(LEVEL_ERROR, "(%s) shutdown failed, err %d\n",
			__FUNCTION__, err);
	}

	err = close(client->fd);
	if (err) {
		err = -errno;
		dprintf(LEVEL_ERROR, "(%s) close failed, err %d\n",
			__FUNCTION__, err);
	}

	gsloop_con_delete_client(gsloop, client);

	free(con);

	return err;
}


static int gsloop_inet_server_close(union gsloop_con	*con)
{
	struct gsloop_server	*server = &con->server;
	struct gsloop		*gsloop = server->gsloop;
	int			err;

	if (server->active_clients != server->attr.max_clients) {
		err = rn_del(&gsloop->rn, server->fd);
		if (err) {
			err = -err;	/* rn returns errno */
			dprintf(LEVEL_ERROR, "(%s) rn_del failed, err %d\n",
				__FUNCTION__, err);
		}
	}

	err = shutdown(server->fd, SHUT_RDWR);
	if (err) {
		err = -errno;
		dprintf(LEVEL_ERROR, "(%s) shutdown failed, err %d\n",
			__FUNCTION__, err);
	}

	err = close(server->fd);
	if (err) {
		err = -errno;
		dprintf(LEVEL_ERROR, "(%s) close failed, err %d\n",
			__FUNCTION__, err);
	}

	gsloop_con_delete_server(gsloop, server);

	free(con);

	return err;
}


static int gsloop_inet_client_detach(struct gsloop_client	*client)
{
	struct gsloop_server	*server = client->server;
	struct gsloop		*gsloop = server->gsloop;
	int			err = 0;

	if (server->active_clients-- == server->attr.max_clients) {
		err = rn_add(&gsloop->rn, server->fd, gsloop_server_callback, server);
		dprintf(LEVEL_DEBUG, "(%s) number of clients (%d) below max (%d), reactivating server\n",
			__PRETTY_FUNCTION__, server->active_clients, server->attr.max_clients);
	}

	return err;
}


int gsloop_tcp_server_create(gsloop_t			*gsloop_anon,
			     uint16_t			port,
			     gsloop_server_attr_t	*attr)
{
	struct gsloop		*gsloop = (struct gsloop *)gsloop_anon;
	struct gsloop_server	*server;
	int			err;
	int			fd;

	if (!gsloop_anon || !attr)
		return -EINVAL;

	if (!attr->client_event && !attr->client_attach)
		return -EINVAL;

	if (attr->max_clients == 0)
		return -EINVAL;

	server = ptx_alloc(struct gsloop_server);
	if (!server) {
		err = -ENOMEM;
		goto exit_0;
	}

	fd = gsloop_inet_server_open(SOCK_STREAM, port);
	if (fd < 0) {
		err = fd;
		goto exit_1;
	}

	server->gsloop		= gsloop;
	server->fd		= fd;
	server->close		= gsloop_inet_server_close;
	server->client_close	= gsloop_inet_client_close;
	server->client_detach	= gsloop_inet_client_detach;
	server->active_clients	= 0;
	memcpy(&server->attr, attr, sizeof(server->attr));
	
	err = rn_prepare_fd_for_add(fd, getpid());
	if (err) {
		err = -err;	/* rn returns errno */
		goto exit_2;
	}

	err = rn_add(&gsloop->rn, fd, gsloop_server_callback, server);
	if (err) {
		err = -err;	/* rn returns errno */
		goto exit_2;
	}

	err = rn_setMask(&gsloop->rn, fd, POLLIN | POLLPRI | POLLERR | POLLHUP);
	if (err) {
		err = -err;     /* rn returns errno */
		goto exit_2;
	}

	gsloop_con_add_server(gsloop, server);
		
	return 0;

 exit_2:
	close(fd);
 exit_1:
	free(server);
 exit_0:
	return err;
}
